from abc import ABC, abstractmethod


class AbstractBaseCommon(ABC):

    @property
    @abstractmethod
    def code(self):
        pass

    @abstractmethod
    def transform(self, **kwargs):
        pass

    @classmethod
    @abstractmethod
    def from_dict(cls, data: dict):
        pass

    # @abstractmethod
    # def to_json(self):
    #     pass


class AbstractBaseField(AbstractBaseCommon, ABC):

    def __init__(self, name, field_type, model=None):
        super().__init__()
        self.name = name
        self.field_type = field_type
        self.model = model

    @property
    def model(self):
        return self.__model

    @model.setter
    def model(self, model):
        self.__model = model

    def transform(self):
        return self.code


class AbstractBaseSerializer(AbstractBaseCommon, ABC):

    def __init__(self, name, field_names, extras=None, model=None, fields=None):
        super().__init__()
        self.name = name
        self.field_names = field_names
        self.extras = extras or {}
        self.__fields = []
        self.model = model

    @property
    def fields(self):

        if not self.__fields:
            for field_name in self.field_names:
                extra = self.extras.get('field_name')
                if extra:
                    field_name = extra.get('source') or field_name
                self.__fields.append(self.model.get_field(field_name))
        return self.__fields

    @property
    def model(self):
        return self.__model

    @model.setter
    def model(self, model):
        self.__model = model
        # TODO: check fields against model at 2020-10-25 by pooria

    def transform(self, **kwargs):
        return self.code


class AbstractBaseModel(AbstractBaseCommon, ABC):

    def __init__(self, name, fields=None, app=None, serializers=None):
        super().__init__()
        self.name = name
        if fields is None:
            self.__fields = {}
        else:
            for field in fields:
                self.register_field(field)
        if serializers is None:
            self.__serializers = {}
        else:
            for serializer in serializers:
                self.register_serializer(serializer)
        self.app = app

    @property
    def fields(self):
        return list(self.__fields.values())

    def get_field(self, name):
        return self.__fields[name]

    def register_field(self, field):
        field.model = self
        self.__fields[field.name] = field

    @property
    def serializers(self):
        return list(self.__serializers.values())

    def get_serializer(self, name):
        return self.__serializers[name]

    def register_serializer(self, serializer):
        # TODO: check serializer fields against self serializer at 2020-10-25 by pooria
        serializer.model = self
        self.__serializers[serializer.name] = serializer

    @property
    def app(self):
        return self.__app

    @app.setter
    def app(self, app):
        self.__app = app

    def transform(self):
        return self.code


# TODO: abstract base endpoint should complete at 2020-10-10 by pooria
class AbstractBaseEndpoint(AbstractBaseCommon, ABC):

    def __init__(self, path, model, serializer, app=None):
        super().__init__()
        self.path = path
        self.__model = model
        self.__serializer = serializer
        self.parent = None
        self.__actions = []
        self.app = app

    @property
    def model(self):
        if not isinstance(self.__model, AbstractBaseModel):
            self.__model = self.app.get_model(self.__model)
        return self.__model

    @property
    def serializer(self):
        if not isinstance(self.__serializer, AbstractBaseSerializer):
            self.__serializer = self.model.get_serializer(self.__serializer)
        return self.__serializer

    @staticmethod
    def needs_router():
        return True

    @abstractmethod
    def assign_app(self, app):
        pass

    @property
    def actions(self):
        return self.__actions

    def register_action(self, action):
        action.parent = self
        self.__actions.append(action)

    @property
    def app(self):
        return self.__app

    @app.setter
    def app(self, app):
        self.__app = app

    def transform(self):
        return self.code


class AbstractBaseApp(AbstractBaseCommon, ABC):

    def __init__(self, name, namespace, service=None, models=None, endpoints=None):
        super().__init__()
        self.name = name
        self.namespace = namespace
        if models is None:
            self.__models = {}
        else:
            for model in models:
                self.register_model(model)
        if endpoints is None:
            self.__endpoints = {}
        else:
            for endpoint in endpoints:
                self.register_endpoint(endpoint)
        self.service = service

    @property
    def models(self):
        return list(self.__models.values())

    def get_model(self, name) -> AbstractBaseModel:
        return self.__models[name]

    def register_model(self, model):
        if model.name in self.__models.keys():
            raise ValueError(f"duplicate model in app: {self.name}, model: {model.name}")
        model.app = self
        self.__models[model.name] = model

    @property
    def endpoints(self):
        return list(self.__endpoints.values())

    def get_endpoint(self, path):
        return self.__endpoints[path]

    def register_endpoint(self, endpoint: AbstractBaseEndpoint):
        endpoint.assign_app(self)
        if endpoint.needs_router():
            self.__endpoints[endpoint.path] = endpoint

        for action in endpoint.actions:
            self.register_endpoint(action)


    @property
    def service(self):
        return self.__service

    @service.setter
    def service(self, service):
        self.__service = service


    @property
    def code(self):
        return "this type of object does not necessarily have a code representation"


class AbstractBaseService(AbstractBaseCommon, ABC):

    def __init__(self, name, deploy_strategy, apps=None):
        super().__init__()
        self.name = name
        self.deploy_strategy = deploy_strategy
        if apps is None:
            self.__apps = {}
        else:
            for app in apps:
                self.register_app(app)

    def register_app(self, app):
        if app.name in self.__apps.keys():
            raise KeyError(f"duplicate app name: {app.name}")
        app.service = self
        self.__apps[app.name] = app

    @property
    def apps(self):
        return list(self.__apps.values())

    def get_app(self, name):
        return self.__apps[name]

    @property
    def code(self):
        return "this type of object does not necessarily have a code representation"




from .abc import AbstractBaseEndpoint, ABC


class SimpleDataAccessEndpoint(AbstractBaseEndpoint, ABC):
    nested_endpoints = {
        "nested_many_simple_data_access": None,
        "nested_foreign_simple_data_access": None
    }

    def assign_app(self, app):
        self.app = app

    @classmethod
    def validate_data(cls, data):
        # TODO: do some validation here at 2020-10-16 by pooria
        pass

    @classmethod
    def from_dict(cls, data: dict):
        cls.validate_data(data)
        path = data.get('path')
        model = data.get('model')
        serializer = data.get('serializer')
        endpoint = cls(model=model, serializer=serializer, path=path)
        actions = data.get('actions', [])
        for action in actions:
            nested_endpoint_type = action['endpoint_type']
            nested_endpoint_class = cls.nested_endpoints[nested_endpoint_type]
            endpoint.register_action(nested_endpoint_class.from_dict(action))
        return endpoint


class NestedManySimpleDataAccessEndpoint(SimpleDataAccessEndpoint, ABC):

    def __init__(self, parent_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__parent_field = parent_field

    @property
    def parent_field(self):
        return self.model.get_field(self.__parent_field)

    @classmethod
    def validate_data(cls, data):
        # TODO: do some validation here at 2020-10-16 by pooria
        pass

    @classmethod
    def from_dict(cls, data: dict):
        cls.validate_data(data)
        path = data.get('path')
        model = data.get('model')
        serializer = data.get('serializer')
        parent_field = data['parent_field']
        endpoint = cls(model=model, parent_field=parent_field, serializer=serializer, path=path)
        return endpoint


class NestedForeignSimpleDataAccessEndpoint(SimpleDataAccessEndpoint, ABC):
    pass


class BaseAddToManyRelationEndpoint(AbstractBaseEndpoint, ABC):

    def __init__(self, parent_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__parent_field = parent_field

    @property
    def parent_field(self):
        return self.parent.model.get_field(self.__parent_field)

    @staticmethod
    def needs_router():
        return False

    def assign_app(self, app):
        self.app = app

    @classmethod
    def validate_data(cls, data: dict):
        return

    @classmethod
    def from_dict(cls, data: dict):
        cls.validate_data(data)
        path = data.get('path')
        model = data.get('model')
        parent_field = data['parent_field']
        endpoint = cls(model=model, parent_field=parent_field, serializer=None, path=path)
        return endpoint


class BaseRemoveFromManyRelationEndpoint(BaseAddToManyRelationEndpoint, ABC):
    pass



from abc import ABC
from .abc import AbstractBaseField


class BaseCharField(AbstractBaseField, ABC):

    def __init__(self, max_length, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_length = max_length

    @classmethod
    def validate_data(cls, data, raise_exception=True):
        assert data.get("field_type", None) == "CharField"
        # TODO: do some validation here at 2020-10-16 by pooria
        pass

    @classmethod
    def from_dict(cls, data: dict):
        cls.validate_data(data, raise_exception=True)
        return cls(**data)


class BaseBooleanField(AbstractBaseField, ABC):

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


class BaseForeignKey(AbstractBaseField, ABC):

    def __init__(self, target, related_name=None, related_query_name=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target = target
        self.related_name = related_name
        self.related_query_name = related_query_name
        self.target_app = None
        self.target_model = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


class BaseManyToMany(AbstractBaseField, ABC):

    def __init__(self, target, related_name=None, related_query_name=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target = target
        self.related_name = related_name
        self.related_query_name = related_query_name
        self.target_app = None
        self.target_model = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


class BaseRelatedField(AbstractBaseField, ABC):

    def __init__(self, target_app, target_model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target_app = target_app
        self.target_model = target_model



from .abc import AbstractBaseSerializer, ABC


class BaseSerializer(AbstractBaseSerializer, ABC):

    @classmethod
    def from_dict(cls, data: dict):
        name = data.get('name')
        fields = data.get('fields')
        extras = data.get('extras')
        return cls(name=name, field_names=fields, extras=extras)



import logging
from collections import defaultdict

from .abc import AbstractBaseApp, AbstractBaseEndpoint, AbstractBaseModel, ABC
from .models import DBModel


logger = logging.getLogger(__name__)


class BaseApp(AbstractBaseApp, ABC):

    _model_choices = {
        "DBModel": None
    }

    _endpoint_choices = {
        "simple_data_access": None
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.relation_targets = defaultdict(list)

    def register_model(self, model):
        super().register_model(model)
        if isinstance(model, DBModel):
            for target, fields in model.relation_targets.items():
                self.relation_targets[target].extend(fields)

    @classmethod
    def validate_data(cls, data, raise_exception):
        # TODO: do some validations here too at 2020-10-16 by pooria
        pass

    @classmethod
    def find_serializer_by_name(cls, serializer_name, models):
        for model in models:
            for serializer in model.serializers:
                if serializer.name == serializer_name:
                    return serializer
        # TODO: raise exception for model not found at 2020-10-16 by pooria

    @classmethod
    def from_dict(cls, data: dict):
        # TODO: make a decision about path at 2020-10-16 by pooria
        cls.validate_data(data, raise_exception=True)
        name = data['name']
        namespace = data['namespace']
        app = cls(name=name, namespace=namespace)
        for model in data['models']:
            logger.info(model)
            model_class = cls._model_choices.get(model.get('model_type', 'default'))
            if model_class is None:
                raise Exception(f"model_type: {model.get('model_type')} is not supported")
            assert issubclass(model_class, AbstractBaseModel)
            app.register_model(model_class.from_dict(model))

        for endpoint in data['endpoints']:
            logger.info(endpoint)
            endpoint_class = cls._endpoint_choices.get(endpoint.get('endpoint_type', 'simple_data_access'))
            if endpoint_class is None:
                raise Exception(f"endpoint_type: {endpoint.get('endpoint_type')} is not supported")
            assert issubclass(endpoint_class, AbstractBaseEndpoint)
            app.register_endpoint(endpoint_class.from_dict(endpoint))

        return app



from collections import defaultdict

from .abc import AbstractBaseService, AbstractBaseApp, ABC
from .apps import BaseApp


class BaseSimpleRestService(AbstractBaseService, ABC):
    # TODO: indicate type of this dictionary should be App at 2020-10-25 by pooria
    _app_choices = {
        "default": None
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.relation_targets = defaultdict(list)

    def register_app(self, app):
        super().register_app(app)
        if isinstance(app, BaseApp):
            for target, fields in app.relation_targets.items():
                self.relation_targets[target].extend(fields)

    def resolve_relation_targets(self):
        for target, fields in self.relation_targets.items():
            app_name, model_name = target.split('.')
            target_app = self.get_app(app_name)
            target_model = target_app.get_model(model_name)
            for field in fields:
                field.target_app = target_app
                field.target_model = target_model

                related_field_class = type(target_model).get_field_type("related_field")
                related_field = related_field_class(name=field.related_name, field_type='related_field',
                                                    target_app=field.model.app, target_model=field.model)
                target_model.register_field(related_field)


    @classmethod
    def from_dict(cls, data: dict):
        apps = data.get('apps')
        deploy_strategy = data.get('deploy_strategy')
        name = data.get('name')
        service = cls(name=name, deploy_strategy=deploy_strategy)

        for app in apps:
            app_class = cls._app_choices.get(app.get('app_type', 'default'))
            if app_class is None:
                raise Exception("app_type not supported")
            assert issubclass(app_class, AbstractBaseApp), "app class should be subclass of AbstractBaseApp"
            service.register_app(app_class.from_dict(app))
        service.resolve_relation_targets()
        return service



import logging
from collections import defaultdict

from .abc import AbstractBaseModel, AbstractBaseSerializer, AbstractBaseField, ABC

from .fields import BaseForeignKey, BaseManyToMany

logger = logging.getLogger(__name__)


class DBModel(AbstractBaseModel, ABC):

    _field_choices = {
        "CharField": None,
        "ForeignKey": None,
        "ManyToMany": None
    }

    _serializer_choices = {
        "default": None
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.relation_targets = defaultdict(list)

    @classmethod
    def get_field_type(cls, field_type):
        return cls._field_choices.get(field_type)

    def register_field(self, field):
        super().register_field(field)
        if isinstance(field, (BaseForeignKey, BaseManyToMany)):
            self.relation_targets[field.target].append(field)

    @classmethod
    def validate_data(cls, data, raise_exception):
        # TODO: do some validations here too at 2020-10-16 by pooria
        pass

    @classmethod
    def from_dict(cls, data: dict):
        cls.validate_data(data, raise_exception=True)
        name = data['name']
        model = cls(name=name)

        for field in data['fields']:
            field_class = cls._field_choices.get(field['field_type'])
            if field_class is None:
                raise Exception(f"field_type: {field['field_type']} is not supported")
            assert issubclass(field_class, AbstractBaseField)
            model.register_field(field_class.from_dict(field))
            logger.debug(field)

        for serializer in data['serializers']:
            # TODO: serializer classes default and stuff at 2020-10-17 by pooria
            serializer_class = cls._serializer_choices.get(serializer.get("serializer_type", "default"))
            if serializer_class is None:
                raise Exception(f"serializer_type: {serializer.get('serializer_type')} is not supported")
            assert issubclass(serializer_class, AbstractBaseSerializer)
            model.register_serializer(serializer_class.from_dict(serializer))

        return model
